from typing import Any, Container


def in_op(a: Container[Any], b: Any) -> bool:
    try:
        return b in a
    except TypeError:
        return False


def not_in_op(a: Container[Any], b: Any) -> bool:
    try:
        return b not in a
    except TypeError:
        return False


def contains_op(a: Any, b: Container[Any]) -> bool:
    try:
        return a in b
    except TypeError:
        return False


def not_contains_op(a: Any, b: Container[Any]) -> bool:
    try:
        return a not in b
    except TypeError:
        return False


def and_op(a: Any, b: Any) -> Any:
    return a and b


def or_op(a: Any, b: Any) -> Any:
    return a or b



from typing import Any, Iterable

from magic_filter.exceptions import RejectOperations, SwitchModeToAll, SwitchModeToAny

from .base import BaseOperation

EMPTY_SLICE = slice(None, None, None)


class GetItemOperation(BaseOperation):
    __slots__ = ("key",)

    def __init__(self, key: Any) -> None:
        self.key = key

    def resolve(self, value: Any, initial_value: Any) -> Any:
        if isinstance(value, Iterable):
            if self.key is ...:
                raise SwitchModeToAny()
            if self.key == EMPTY_SLICE:
                raise SwitchModeToAll(self.key)
        try:
            return value[self.key]
        except (KeyError, IndexError, TypeError) as e:
            raise RejectOperations(e) from e



from typing import TYPE_CHECKING, Any, Iterable

from magic_filter.operations import BaseOperation

if TYPE_CHECKING:
    from magic_filter.magic import MagicFilter


class ExtractOperation(BaseOperation):
    __slots__ = ("extractor",)

    def __init__(self, extractor: "MagicFilter") -> None:
        self.extractor = extractor

    def resolve(self, value: Any, initial_value: Any) -> Any:
        if not isinstance(value, Iterable):
            return None

        result = []
        for item in value:
            if self.extractor.resolve(item):
                result.append(item)
        return result



from typing import TYPE_CHECKING, Any

from magic_filter.exceptions import RejectOperations
from magic_filter.operations import BaseOperation

if TYPE_CHECKING:
    from magic_filter import MagicFilter


class SelectorOperation(BaseOperation):
    __slots__ = ("inner",)

    def __init__(self, inner: "MagicFilter"):
        self.inner = inner

    def resolve(self, value: Any, initial_value: Any) -> Any:
        if self.inner.resolve(value):
            return value
        raise RejectOperations()



from typing import Any, Callable

from ..helper import resolve_if_needed
from .base import BaseOperation


class ComparatorOperation(BaseOperation):
    __slots__ = (
        "right",
        "comparator",
    )

    def __init__(self, right: Any, comparator: Callable[[Any, Any], bool]) -> None:
        self.right = right
        self.comparator = comparator

    def resolve(self, value: Any, initial_value: Any) -> Any:
        return self.comparator(value, resolve_if_needed(self.right, initial_value=initial_value))



from . import operations
from .attrdict import AttrDict
from .magic import MagicFilter, MagicT, RegexpMode

__all__ = (
    "__version__",
    "operations",
    "MagicFilter",
    "MagicT",
    "RegexpMode",
    "F",
    "AttrDict",
)

__version__ = "1"

F = MagicFilter()



from abc import ABC, abstractmethod
from typing import Any


class BaseOperation(ABC):
    important: bool = False

    @abstractmethod
    def resolve(self, value: Any, initial_value: Any) -> Any:  # pragma: no cover
        pass



from typing import Any, Dict, TypeVar

KT = TypeVar("KT")
VT = TypeVar("VT")


class AttrDict(Dict[KT, VT]):
    """
    A wrapper over dict which where element can be accessed as regular attributes
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super(AttrDict, self).__init__(*args, **kwargs)
        self.__dict__ = self  # type: ignore



from typing import Any, Dict, Tuple

from ..exceptions import RejectOperations
from .base import BaseOperation


class CallOperation(BaseOperation):
    __slots__ = ("args", "kwargs")

    def __init__(self, args: Tuple[Any, ...], kwargs: Dict[str, Any]):
        self.args = args
        self.kwargs = kwargs

    def resolve(self, value: Any, initial_value: Any) -> Any:
        if not callable(value):
            raise RejectOperations(TypeError(f"{value} is not callable"))
        return value(*self.args, **self.kwargs)



import operator
import re
from functools import wraps
from typing import Any, Callable, Container, Optional, Pattern, Tuple, Type, TypeVar, Union
from warnings import warn

from magic_filter.exceptions import (
    ParamsConflict,
    RejectOperations,
    SwitchModeToAll,
    SwitchModeToAny,
)
from magic_filter.operations import (
    BaseOperation,
    CallOperation,
    CastOperation,
    CombinationOperation,
    ComparatorOperation,
    ExtractOperation,
    FunctionOperation,
    GetAttributeOperation,
    GetItemOperation,
    ImportantCombinationOperation,
    ImportantFunctionOperation,
    RCombinationOperation,
    SelectorOperation,
)
from magic_filter.util import and_op, contains_op, in_op, not_contains_op, not_in_op, or_op

MagicT = TypeVar("MagicT", bound="MagicFilter")


class RegexpMode:
    SEARCH = "search"
    MATCH = "match"
    FINDALL = "findall"
    FINDITER = "finditer"
    FULLMATCH = "fullmatch"


class MagicFilter:
    __slots__ = ("_operations",)

    def __init__(self, operations: Tuple[BaseOperation, ...] = ()) -> None:
        self._operations = operations

    # An instance of MagicFilter cannot be used as an iterable object because objects
    # with a __getitem__ method can be endlessly iterated, which is not the desired behavior.
    __iter__ = None

    @classmethod
    def ilter(cls, magic: "MagicFilter") -> Callable[[Any], Any]:
        @wraps(magic.resolve)
        def wrapper(value: Any) -> Any:
            return magic.resolve(value)

        return wrapper

    @classmethod
    def _new(cls: Type[MagicT], operations: Tuple[BaseOperation, ...]) -> MagicT:
        return cls(operations=operations)

    def _extend(self: MagicT, operation: BaseOperation) -> MagicT:
        return self._new(self._operations + (operation,))

    def _replace_last(self: MagicT, operation: BaseOperation) -> MagicT:
        return self._new(self._operations[:-1] + (operation,))

    def _exclude_last(self: MagicT) -> MagicT:
        return self._new(self._operations[:-1])

    def _resolve(self, value: Any, operations: Optional[Tuple[BaseOperation, ...]] = None) -> Any:
        initial_value = value
        if operations is None:
            operations = self._operations
        rejected = False
        for index, operation in enumerate(operations):
            if rejected and not operation.important:
                continue
            try:
                value = operation.resolve(value=value, initial_value=initial_value)
            except SwitchModeToAll:
                return all(self._resolve(value=item, operations=operations[index + 1 :]) for item in value)
            except SwitchModeToAny:
                return any(self._resolve(value=item, operations=operations[index + 1 :]) for item in value)
            except RejectOperations:
                rejected = True
                value = None
                continue
            rejected = False
        return value

    def __bool__(self) -> bool:
        return True

    def resolve(self: MagicT, value: Any) -> Any:
        return self._resolve(value=value)

    def __getattr__(self: MagicT, item: Any) -> MagicT:
        if item.startswith("_"):
            raise AttributeError(f"{type(self).__name__!r} object has no attribute {item!r}")
        return self._extend(GetAttributeOperation(name=item))

    attr_ = __getattr__

    def __getitem__(self: MagicT, item: Any) -> MagicT:
        if isinstance(item, MagicFilter):
            return self._extend(SelectorOperation(inner=item))
        return self._extend(GetItemOperation(key=item))

    def __len__(self) -> int:
        raise TypeError(f"Length can't be taken using len() function. Use {type(self).__name__}.len() instead.")

    def __eq__(self: MagicT, other: Any) -> MagicT:  # type: ignore
        return self._extend(ComparatorOperation(right=other, comparator=operator.eq))

    def __ne__(self: MagicT, other: Any) -> MagicT:  # type: ignore
        return self._extend(ComparatorOperation(right=other, comparator=operator.ne))

    def __lt__(self: MagicT, other: Any) -> MagicT:
        return self._extend(ComparatorOperation(right=other, comparator=operator.lt))

    def __gt__(self: MagicT, other: Any) -> MagicT:
        return self._extend(ComparatorOperation(right=other, comparator=operator.gt))

    def __le__(self: MagicT, other: Any) -> MagicT:
        return self._extend(ComparatorOperation(right=other, comparator=operator.le))

    def __ge__(self: MagicT, other: Any) -> MagicT:
        return self._extend(ComparatorOperation(right=other, comparator=operator.ge))

    def __invert__(self: MagicT) -> MagicT:
        if (
            self._operations
            and isinstance(self._operations[-1], ImportantFunctionOperation)
            and self._operations[-1].function == operator.not_
        ):
            return self._exclude_last()
        return self._extend(ImportantFunctionOperation(function=operator.not_))

    def __call__(self: MagicT, *args: Any, **kwargs: Any) -> MagicT:
        return self._extend(CallOperation(args=args, kwargs=kwargs))

    def __and__(self: MagicT, other: Any) -> MagicT:
        if isinstance(other, MagicFilter):
            return self._extend(CombinationOperation(right=other, combinator=and_op))
        return self._extend(CombinationOperation(right=other, combinator=operator.and_))

    def __rand__(self: MagicT, other: Any) -> MagicT:
        return self._extend(RCombinationOperation(left=other, combinator=operator.and_))

    def __or__(self: MagicT, other: Any) -> MagicT:
        if isinstance(other, MagicFilter):
            return self._extend(ImportantCombinationOperation(right=other, combinator=or_op))
        return self._extend(ImportantCombinationOperation(right=other, combinator=operator.or_))

    def __ror__(self: MagicT, other: Any) -> MagicT:
        return self._extend(RCombinationOperation(left=other, combinator=operator.or_))

    def __xor__(self: MagicT, other: Any) -> MagicT:
        return self._extend(CombinationOperation(right=other, combinator=operator.xor))

    def __rxor__(self: MagicT, other: Any) -> MagicT:
        return self._extend(RCombinationOperation(left=other, combinator=operator.xor))

    def __rshift__(self: MagicT, other: Any) -> MagicT:
        return self._extend(CombinationOperation(right=other, combinator=operator.rshift))

    def __rrshift__(self: MagicT, other: Any) -> MagicT:
        return self._extend(RCombinationOperation(left=other, combinator=operator.rshift))

    def __lshift__(self: MagicT, other: Any) -> MagicT:
        return self._extend(CombinationOperation(right=other, combinator=operator.lshift))

    def __rlshift__(self: MagicT, other: Any) -> MagicT:
        return self._extend(RCombinationOperation(left=other, combinator=operator.lshift))

    def __add__(self: MagicT, other: Any) -> MagicT:
        return self._extend(CombinationOperation(right=other, combinator=operator.add))

    def __radd__(self: MagicT, other: Any) -> MagicT:
        return self._extend(RCombinationOperation(left=other, combinator=operator.add))

    def __sub__(self: MagicT, other: Any) -> MagicT:
        return self._extend(CombinationOperation(right=other, combinator=operator.sub))

    def __rsub__(self: MagicT, other: Any) -> MagicT:
        return self._extend(RCombinationOperation(left=other, combinator=operator.sub))

    def __mul__(self: MagicT, other: Any) -> MagicT:
        return self._extend(CombinationOperation(right=other, combinator=operator.mul))

    def __rmul__(self: MagicT, other: Any) -> MagicT:
        return self._extend(RCombinationOperation(left=other, combinator=operator.mul))

    def __truediv__(self: MagicT, other: Any) -> MagicT:
        return self._extend(CombinationOperation(right=other, combinator=operator.truediv))

    def __rtruediv__(self: MagicT, other: Any) -> MagicT:
        return self._extend(RCombinationOperation(left=other, combinator=operator.truediv))

    def __floordiv__(self: MagicT, other: Any) -> MagicT:
        return self._extend(CombinationOperation(right=other, combinator=operator.floordiv))

    def __rfloordiv__(self: MagicT, other: Any) -> MagicT:
        return self._extend(RCombinationOperation(left=other, combinator=operator.floordiv))

    def __mod__(self: MagicT, other: Any) -> MagicT:
        return self._extend(CombinationOperation(right=other, combinator=operator.mod))

    def __rmod__(self: MagicT, other: Any) -> MagicT:
        return self._extend(RCombinationOperation(left=other, combinator=operator.mod))

    def __matmul__(self: MagicT, other: Any) -> MagicT:
        return self._extend(CombinationOperation(right=other, combinator=operator.matmul))

    def __rmatmul__(self: MagicT, other: Any) -> MagicT:
        return self._extend(RCombinationOperation(left=other, combinator=operator.matmul))

    def __pow__(self: MagicT, other: Any) -> MagicT:
        return self._extend(CombinationOperation(right=other, combinator=operator.pow))

    def __rpow__(self: MagicT, other: Any) -> MagicT:
        return self._extend(RCombinationOperation(left=other, combinator=operator.pow))

    def __pos__(self: MagicT) -> MagicT:
        return self._extend(FunctionOperation(function=operator.pos))

    def __neg__(self: MagicT) -> MagicT:
        return self._extend(FunctionOperation(function=operator.neg))

    def is_(self: MagicT, value: Any) -> MagicT:
        return self._extend(CombinationOperation(right=value, combinator=operator.is_))

    def is_not(self: MagicT, value: Any) -> MagicT:
        return self._extend(CombinationOperation(right=value, combinator=operator.is_not))

    def in_(self: MagicT, iterable: Union[Container[Any], MagicT]) -> MagicT:
        return self._extend(FunctionOperation(in_op, iterable))

    def not_in(self: MagicT, iterable: Union[Container[Any], MagicT]) -> MagicT:
        return self._extend(FunctionOperation(not_in_op, iterable))

    def contains(self: MagicT, value: Any) -> MagicT:
        return self._extend(FunctionOperation(contains_op, value))

    def not_contains(self: MagicT, value: Any) -> MagicT:
        return self._extend(FunctionOperation(not_contains_op, value))

    def len(self: MagicT) -> MagicT:
        return self._extend(FunctionOperation(len))

    def regexp(
        self: MagicT,
        pattern: Union[str, Pattern[str]],
        *,
        mode: Optional[str] = None,
        search: Optional[bool] = None,
        flags: Union[int, re.RegexFlag] = 0,
    ) -> MagicT:

        if search is not None:
            warn(
                "Param 'search' is deprecated, use 'mode' instead.",
                DeprecationWarning,
            )

            if mode is not None:
                msg = "Can't pass both 'search' and 'mode' params."
                raise ParamsConflict(msg)

            mode = RegexpMode.SEARCH if search else RegexpMode.MATCH

        if mode is None:
            mode = RegexpMode.MATCH

        if isinstance(pattern, str):
            pattern = re.compile(pattern, flags=flags)

        regexp_func = getattr(pattern, mode)
        return self._extend(FunctionOperation(regexp_func))

    def func(self: MagicT, func: Callable[[Any], Any], *args: Any, **kwargs: Any) -> MagicT:
        return self._extend(FunctionOperation(func, *args, **kwargs))

    def cast(self: MagicT, func: Callable[[Any], Any]) -> MagicT:
        return self._extend(CastOperation(func))

    def extract(self: MagicT, magic: "MagicT") -> MagicT:
        return self._extend(ExtractOperation(magic))



from typing import Any, Callable

from ..exceptions import RejectOperations
from ..helper import resolve_if_needed
from .base import BaseOperation


class FunctionOperation(BaseOperation):
    __slots__ = (
        "function",
        "args",
        "kwargs",
    )

    def __init__(self, function: Callable[..., Any], *args: Any, **kwargs: Any) -> None:
        self.function = function
        self.args = args
        self.kwargs = kwargs

    def resolve(self, value: Any, initial_value: Any) -> Any:
        try:
            return self.function(
                *(resolve_if_needed(arg, initial_value) for arg in self.args),
                value,
                **{key: resolve_if_needed(value, initial_value) for key, value in self.kwargs.items()},
            )
        except (TypeError, ValueError) as e:
            raise RejectOperations(e) from e


class ImportantFunctionOperation(FunctionOperation):
    important = True



class MagicFilterException(Exception):
    pass


class SwitchMode(MagicFilterException):
    pass


class SwitchModeToAll(SwitchMode):
    def __init__(self, key: slice) -> None:
        self.key = key


class SwitchModeToAny(SwitchMode):
    pass


class RejectOperations(MagicFilterException):
    pass


class ParamsConflict(MagicFilterException):
    pass



from .base import BaseOperation
from .call import CallOperation
from .cast import CastOperation
from .combination import CombinationOperation, ImportantCombinationOperation, RCombinationOperation
from .comparator import ComparatorOperation
from .extract import ExtractOperation
from .function import FunctionOperation, ImportantFunctionOperation
from .getattr import GetAttributeOperation
from .getitem import GetItemOperation
from .selector import SelectorOperation

__all__ = (
    "BaseOperation",
    "CallOperation",
    "CastOperation",
    "CombinationOperation",
    "ComparatorOperation",
    "FunctionOperation",
    "GetAttributeOperation",
    "GetItemOperation",
    "ImportantCombinationOperation",
    "ImportantFunctionOperation",
    "RCombinationOperation",
    "SelectorOperation",
    "ExtractOperation",
)



from typing import Any


def resolve_if_needed(value: Any, initial_value: Any) -> Any:
    # To avoid circular imports here is used local import
    from magic_filter import MagicFilter

    if not isinstance(value, MagicFilter):
        return value
    return value.resolve(initial_value)



from abc import ABC
from typing import Any

from ..exceptions import RejectOperations
from .base import BaseOperation


class GetAttributeOperation(BaseOperation, ABC):
    __slots__ = ("name",)

    def __init__(self, name: str) -> None:
        self.name = name

    def resolve(self, value: Any, initial_value: Any) -> Any:
        try:
            return getattr(value, self.name)
        except AttributeError as e:
            raise RejectOperations(e) from e



from typing import Any, Callable

from ..helper import resolve_if_needed
from .base import BaseOperation


class CombinationOperation(BaseOperation):
    __slots__ = (
        "right",
        "combinator",
    )

    def __init__(self, right: Any, combinator: Callable[[Any, Any], bool]) -> None:
        self.right = right
        self.combinator = combinator

    def resolve(self, value: Any, initial_value: Any) -> Any:
        return self.combinator(value, resolve_if_needed(self.right, initial_value=initial_value))


class ImportantCombinationOperation(CombinationOperation):
    important = True


class RCombinationOperation(BaseOperation):
    __slots__ = (
        "left",
        "combinator",
    )

    def __init__(self, left: Any, combinator: Callable[[Any, Any], bool]) -> None:
        self.left = left
        self.combinator = combinator

    def resolve(self, value: Any, initial_value: Any) -> Any:
        return self.combinator(resolve_if_needed(self.left, initial_value), value)



from typing import Any, Callable

from ..exceptions import RejectOperations
from .base import BaseOperation


class CastOperation(BaseOperation):
    __slots__ = ("func",)

    def __init__(self, func: Callable[[Any], Any]) -> None:
        self.func = func

    def resolve(self, value: Any, initial_value: Any) -> Any:
        try:
            return self.func(value)
        except Exception as e:
            raise RejectOperations(e) from e
